@@ -1,8 +1,8 @@
 # Copyright (C) 2016 JWCrypto Project Contributors - see LICENSE file
 
-import abc
 import os
 import struct
+from abc import ABCMeta, abstractmethod
 from binascii import hexlify, unhexlify
 
 from cryptography.exceptions import InvalidSignature
@@ -17,8 +17,6 @@ from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
 from cryptography.hazmat.primitives.keywrap import aes_key_unwrap, aes_key_wrap
 from cryptography.hazmat.primitives.padding import PKCS7
 
-import six
-
 from jwcrypto.common import InvalidCEKeyLength
 from jwcrypto.common import InvalidJWAAlgorithm
 from jwcrypto.common import InvalidJWEKeyLength
@@ -31,33 +29,32 @@ from jwcrypto.jwk import JWK
 # Implements RFC 7518 - JSON Web Algorithms (JWA)
 
 
-@six.add_metaclass(abc.ABCMeta)
-class JWAAlgorithm(object):
+class JWAAlgorithm(metaclass=ABCMeta):
 
-    @abc.abstractproperty
+    @property
+    @abstractmethod
     def name(self):
         """The algorithm Name"""
-        pass
 
-    @abc.abstractproperty
+    @property
+    @abstractmethod
     def description(self):
         """A short description"""
-        pass
 
-    @abc.abstractproperty
+    @property
+    @abstractmethod
     def keysize(self):
         """The actual/recommended/minimum key size"""
-        pass
 
-    @abc.abstractproperty
+    @property
+    @abstractmethod
     def algorithm_usage_location(self):
         """One of 'alg', 'enc' or 'JWK'"""
-        pass
 
-    @abc.abstractproperty
+    @property
+    @abstractmethod
     def algorithm_use(self):
         """One of 'sig', 'kex', 'enc'"""
-        pass
 
 
 def _bitsize(x):
@@ -85,7 +82,7 @@ def _decode_int(n):
     return int(hexlify(n), 16)
 
 
-class _RawJWS(object):
+class _RawJWS:
 
     def sign(self, key, payload):
         raise NotImplementedError
@@ -141,9 +138,9 @@ class _RawEC(_RawJWS):
 
     def sign(self, key, payload):
         skey = key.get_op_key('sign', self._curve)
+        size = skey.key_size
         signature = skey.sign(payload, ec.ECDSA(self.hashfn))
         r, s = ec_utils.decode_dss_signature(signature)
-        size = key.get_curve(self._curve).key_size
         return _encode_int(r, size) + _encode_int(s, size)
 
     def verify(self, key, payload, signature):
@@ -336,7 +333,7 @@ class _None(_RawNone, JWAAlgorithm):
     algorithm_use = 'sig'
 
 
-class _RawKeyMgmt(object):
+class _RawKeyMgmt:
 
     def wrap(self, key, bitsize, cek, headers):
         raise NotImplementedError
@@ -356,7 +353,7 @@ class _RSA(_RawKeyMgmt):
         if key['kty'] != 'RSA':
             raise InvalidJWEKeyType('RSA', key['kty'])
 
-    # FIXME: get key size and insure > 2048 bits
+    # FIXME: get key size and ensure > 2048 bits
     def wrap(self, key, bitsize, cek, headers):
         self._check_key(key)
         if not cek:
@@ -367,7 +364,7 @@ class _RSA(_RawKeyMgmt):
 
     def unwrap(self, key, bitsize, ek, headers):
         self._check_key(key)
-        rk = key.get_op_key('decrypt')
+        rk = key.get_op_key('unwrapKey')
         cek = rk.decrypt(ek, self.padfn)
         if _bitsize(cek) != bitsize:
             raise InvalidJWEKeyLength(bitsize, _bitsize(cek))
@@ -584,7 +581,7 @@ class _Pbes2HsAesKw(_RawKeyMgmt):
 
     def _get_key(self, alg, key, p2s, p2c):
         if not isinstance(key, JWK):
-            # backwards compatiblity for old interface
+            # backwards compatibility for old interface
             if isinstance(key, bytes):
                 plain = key
             else:
@@ -839,7 +836,6 @@ class _EdDsa(_RawJWS, JWAAlgorithm):
     keysize = None
 
     def sign(self, key, payload):
-
         if key['crv'] in ['Ed25519', 'Ed448']:
             skey = key.get_op_key('sign')
             return skey.sign(payload)
@@ -852,7 +848,7 @@ class _EdDsa(_RawJWS, JWAAlgorithm):
         raise NotImplementedError
 
 
-class _RawJWE(object):
+class _RawJWE:
 
     def encrypt(self, k, a, m):
         raise NotImplementedError
@@ -1045,7 +1041,7 @@ class _A256Gcm(_AesGcm, JWAAlgorithm):
     algorithm_use = 'enc'
 
 
-class JWA(object):
+class JWA:
     """JWA Signing Algorithms.
 
     This class provides access to all JWA algorithms.
@@ -1104,21 +1100,21 @@ class JWA(object):
         try:
             return cls.instantiate_alg(name, use='sig')
         except KeyError:
-            raise InvalidJWAAlgorithm(
-                '%s is not a valid Signign algorithm name' % name)
+            raise InvalidJWAAlgorithm('%s is not a valid Signing algorithm'
+                                      ' name' % name) from None
 
     @classmethod
     def keymgmt_alg(cls, name):
         try:
             return cls.instantiate_alg(name, use='kex')
         except KeyError:
-            raise InvalidJWAAlgorithm(
-                '%s is not a valid Key Management algorithm name' % name)
+            raise InvalidJWAAlgorithm('%s is not a valid Key Management'
+                                      ' algorithm name' % name) from None
 
     @classmethod
     def encryption_alg(cls, name):
         try:
             return cls.instantiate_alg(name, use='enc')
         except KeyError:
-            raise InvalidJWAAlgorithm(
-                '%s is not a valid Encryption algorithm name' % name)
+            raise InvalidJWAAlgorithm('%s is not a valid Encryption'
+                                      ' algorithm name' % name) from None
