@@ -13,14 +13,12 @@ from cryptography.hazmat.primitives.asymmetric import rsa
 
 from deprecated import deprecated
 
-from six import iteritems
-
 from jwcrypto.common import JWException
 from jwcrypto.common import base64url_decode, base64url_encode
 from jwcrypto.common import json_decode, json_encode
 
 
-class UnimplementedOKPCurveKey(object):
+class UnimplementedOKPCurveKey:
     @classmethod
     def generate(cls):
         raise NotImplementedError
@@ -76,12 +74,15 @@ except ImportError:
     X448PrivateKey = UnimplementedOKPCurveKey
 
 
-_OKP_CURVE = namedtuple('Name', 'pubkey privkey')
+_Ed25519_CURVE = namedtuple('Ed25519', 'pubkey privkey')
+_Ed448_CURVE = namedtuple('Ed448', 'pubkey privkey')
+_X25519_CURVE = namedtuple('X25519', 'pubkey privkey')
+_X448_CURVE = namedtuple('X448', 'pubkey privkey')
 _OKP_CURVES_TABLE = {
-    'Ed25519': _OKP_CURVE(Ed25519PublicKey, Ed25519PrivateKey),
-    'Ed448': _OKP_CURVE(Ed448PublicKey, Ed448PrivateKey),
-    'X25519': _OKP_CURVE(X25519PublicKey, X25519PrivateKey),
-    'X448': _OKP_CURVE(X448PublicKey, X448PrivateKey)
+    'Ed25519': _Ed25519_CURVE(Ed25519PublicKey, Ed25519PrivateKey),
+    'Ed448': _Ed448_CURVE(Ed448PublicKey, Ed448PrivateKey),
+    'X25519': _X25519_CURVE(X25519PublicKey, X25519PrivateKey),
+    'X448': _X448_CURVE(X448PublicKey, X448PrivateKey)
 }
 
 
@@ -245,7 +246,7 @@ class InvalidJWKOperation(JWException):
             op = JWKOperationsRegistry[self.op]
         else:
             op = 'Unknown(%s)' % self.op
-        valid = list()
+        valid = []
         for v in self.values:
             if v in list(JWKOperationsRegistry.keys()):
                 valid.append(JWKOperationsRegistry[v])
@@ -263,13 +264,11 @@ class InvalidJWKValue(JWException):
     on the key type or other constraints.
     """
 
-    pass
-
 
 class JWK(dict):
     """JSON Web Key object
 
-    This object represent a Key.
+    This object represents a Key.
     It must be instantiated by using the standard defined key/value pairs
     as arguments of the initialization function.
     """
@@ -296,7 +295,7 @@ class JWK(dict):
          * OKP: crv(str) (one of Ed25519, Ed448, X25519, X448)
 
         Deprecated:
-        Alternatively if the 'generate' parameter is provided, with a
+        Alternatively if the 'generate' parameter is provided with a
         valid key type as value then a new key will be generated according
         to the defaults or provided key strength options (type specific).
 
@@ -305,6 +304,8 @@ class JWK(dict):
             are provided.
         """
         super(JWK, self).__init__()
+        self._cache_pub_k = None
+        self._cache_pri_k = None
 
         if 'generate' in kwargs:
             self.generate_key(**kwargs)
@@ -318,8 +319,8 @@ class JWK(dict):
         try:
             kty = kwargs['kty']
             gen = getattr(obj, '_generate_%s' % kty)
-        except (KeyError, AttributeError):
-            raise InvalidJWKType(kty)
+        except (KeyError, AttributeError) as e:
+            raise InvalidJWKType(kty) from e
         gen(kwargs)
         return obj
 
@@ -328,8 +329,8 @@ class JWK(dict):
         try:
             kty = params.pop('generate')
             gen = getattr(self, '_generate_%s' % kty)
-        except (KeyError, AttributeError):
-            raise InvalidJWKType(kty)
+        except (KeyError, AttributeError) as e:
+            raise InvalidJWKType(kty) from e
 
         gen(params)
 
@@ -341,8 +342,8 @@ class JWK(dict):
             try:
                 from jwcrypto.jwa import JWA
                 alg = JWA.instantiate_alg(params['alg'])
-            except KeyError:
-                raise ValueError("Invalid 'alg' parameter")
+            except KeyError as e:
+                raise ValueError("Invalid 'alg' parameter") from e
             size = alg.keysize
         return size
 
@@ -397,19 +398,45 @@ class JWK(dict):
         )
         self.import_key(**params)
 
-    def _get_curve_by_name(self, name):
-        if name == 'P-256':
-            return ec.SECP256R1()
-        elif name == 'P-384':
-            return ec.SECP384R1()
-        elif name == 'P-521':
-            return ec.SECP521R1()
-        elif name == 'secp256k1':
-            return ec.SECP256K1()
-        elif name in _OKP_CURVES_TABLE:
-            return name
+    def _get_curve_by_name(self, name, ctype=None):
+        crv = self.get('crv')
+
+        if name is None:
+            cname = crv
+        elif name == 'P-256K':
+            # P-256K is an alias for 'secp256k1' to handle compatibility
+            # with some implementation using this old drafting name
+            cname = 'secp256k1'
         else:
-            raise InvalidJWKValue('Unknown Elliptic Curve Type')
+            cname = name
+
+        # Check we are asking for the correct curve unless this is being
+        # requested for generation on a blank JWK object
+        if crv:
+            ccrv = crv
+            if ccrv == 'P-256K':
+                ccrv = 'secp256k1'
+            if ccrv != cname:
+                raise InvalidJWKValue('Curve requested is "%s", but '
+                                      'key curve is "%s"' % (name, crv))
+        kty = self.get('kty')
+        if kty is not None and ctype is not None and kty != ctype:
+            raise InvalidJWKType('Curve Requested is of type "%s", but '
+                                 'key curve is of type "%s"' % (ctype, kty))
+
+        # Return a curve object
+        if cname == 'P-256':
+            return ec.SECP256R1()
+        elif cname == 'P-384':
+            return ec.SECP384R1()
+        elif cname == 'P-521':
+            return ec.SECP521R1()
+        elif cname == 'secp256k1':
+            return ec.SECP256K1()
+        elif cname in _OKP_CURVES_TABLE:
+            return _OKP_CURVES_TABLE[cname]
+        else:
+            raise InvalidJWKValue('Unknown Curve Name [%s]' % (name))
 
     def _generate_EC(self, params):
         curve = 'P-256'
@@ -419,8 +446,8 @@ class JWK(dict):
         # precedence
         if 'crv' in params:
             curve = params.pop('crv')
-        curve_name = self._get_curve_by_name(curve)
-        key = ec.generate_private_key(curve_name, default_backend())
+        curve_fn = self._get_curve_by_name(curve, 'EC')
+        key = ec.generate_private_key(curve_fn, default_backend())
         self._import_pyca_pri_ec(key, **params)
 
     def _import_pyca_pri_ec(self, key, **params):
@@ -449,15 +476,12 @@ class JWK(dict):
     def _generate_OKP(self, params):
         if 'crv' not in params:
             raise InvalidJWKValue('Must specify "crv" for OKP key generation')
-        try:
-            key = _OKP_CURVES_TABLE[params['crv']].privkey.generate()
-        except KeyError:
-            raise InvalidJWKValue('"%s" is not a supported curve for the '
-                                  'OKP key type' % params['crv'])
+        curve_fn = self._get_curve_by_name(params['crv'], 'OKP')
+        key = curve_fn.privkey.generate()
         self._import_pyca_pri_okp(key, **params)
 
     def _okp_curve_from_pyca_key(self, key):
-        for name, val in iteritems(_OKP_CURVES_TABLE):
+        for name, val in _OKP_CURVES_TABLE.items():
             if isinstance(key, (val.pubkey, val.privkey)):
                 return name
         raise InvalidJWKValue('Invalid OKP Key object %r' % key)
@@ -487,8 +511,10 @@ class JWK(dict):
         self.import_key(**params)
 
     def import_key(self, **kwargs):
-        newkey = dict()
+        newkey = {}
         key_vals = 0
+        self._cache_pub_k = None
+        self._cache_pri_k = None
 
         names = list(kwargs.keys())
 
@@ -509,7 +535,7 @@ class JWK(dict):
                 while name in names:
                     names.remove(name)
 
-        for name, val in iteritems(JWKValuesRegistry[kty]):
+        for name, val in JWKValuesRegistry[kty].items():
             if val.required and name not in newkey:
                 raise InvalidJWKValue('Missing required value %s' % name)
             if val.type == ParmType.unsupported and name in newkey:
@@ -518,18 +544,18 @@ class JWK(dict):
                 # Check that the value is base64url encoded
                 try:
                     base64url_decode(newkey[name])
-                except Exception:  # pylint: disable=broad-except
+                except Exception as e:  # pylint: disable=broad-except
                     raise InvalidJWKValue(
                         '"%s" is not base64url encoded' % name
-                    )
+                    ) from e
             if val.type == ParmType.b64u and name in newkey:
                 # Check that the value is Base64urlUInt encoded
                 try:
                     self._decode_int(newkey[name])
-                except Exception:  # pylint: disable=broad-except
+                except Exception as e:  # pylint: disable=broad-except
                     raise InvalidJWKValue(
                         '"%s" is not Base64urlUInt encoded' % name
-                    )
+                    ) from e
 
         # Unknown key parameters are allowed
         for name in names:
@@ -581,7 +607,7 @@ class JWK(dict):
         try:
             jkey = json_decode(key)
         except Exception as e:  # pylint: disable=broad-except
-            raise InvalidJWKValue(e)
+            raise InvalidJWKValue from e
         obj.import_key(**jkey)
         return obj
 
@@ -595,7 +621,7 @@ class JWK(dict):
         """
         if private_key is True:
             # Use _export_all for backwards compatibility, as this
-            # function allows to export symmetrict keys too
+            # function allows to export symmetric keys too
             return self._export_all(as_dict)
 
         return self.export_public(as_dict)
@@ -628,7 +654,7 @@ class JWK(dict):
         return pub
 
     def _export_all(self, as_dict=False):
-        d = dict()
+        d = {}
         d.update(self)
         if as_dict is True:
             return d
@@ -702,22 +728,16 @@ class JWK(dict):
             raise InvalidJWKType('Not an EC or OKP key')
         return self.get('crv')
 
+    @deprecated
     def get_curve(self, arg):
         """Gets the Elliptic Curve associated with the key.
 
         :param arg: an optional curve name
 
         :raises InvalidJWKType: the key is not an EC or OKP key.
-        :raises InvalidJWKValue: if the curve names is invalid.
+        :raises InvalidJWKValue: if the curve name is invalid.
         """
-        crv = self.get('crv')
-        if self.get('kty') not in ['EC', 'OKP']:
-            raise InvalidJWKType('Not an EC or OKP key')
-        if arg and crv != arg:
-            raise InvalidJWKValue('Curve requested is "%s", but '
-                                  'key curve is "%s"' % (arg, crv))
-
-        return self._get_curve_by_name(crv)
+        return self._get_curve_by_name(arg)
 
     def _check_constraints(self, usage, operation):
         use = self.get('use')
@@ -734,57 +754,94 @@ class JWK(dict):
     def _decode_int(self, n):
         return int(hexlify(base64url_decode(n)), 16)
 
-    def _rsa_pub(self):
+    def _rsa_pub_n(self):
         e = self._decode_int(self.get('e'))
         n = self._decode_int(self.get('n'))
         return rsa.RSAPublicNumbers(e, n)
 
-    def _rsa_pri(self):
+    def _rsa_pri_n(self):
         p = self._decode_int(self.get('p'))
         q = self._decode_int(self.get('q'))
         d = self._decode_int(self.get('d'))
         dp = self._decode_int(self.get('dp'))
         dq = self._decode_int(self.get('dq'))
         qi = self._decode_int(self.get('qi'))
-        return rsa.RSAPrivateNumbers(p, q, d, dp, dq, qi, self._rsa_pub())
+        return rsa.RSAPrivateNumbers(p, q, d, dp, dq, qi, self._rsa_pub_n())
 
-    def _ec_pub(self, curve):
+    def _rsa_pub(self):
+        k = self._cache_pub_k
+        if k is None:
+            k = self._rsa_pub_n().public_key(default_backend())
+            self._cache_pub_k = k
+        return k
+
+    def _rsa_pri(self):
+        k = self._cache_pri_k
+        if k is None:
+            k = self._rsa_pri_n().private_key(default_backend())
+            self._cache_pri_k = k
+        return k
+
+    def _ec_pub_n(self, curve):
         x = self._decode_int(self.get('x'))
         y = self._decode_int(self.get('y'))
-        return ec.EllipticCurvePublicNumbers(x, y, self.get_curve(curve))
+        curve_fn = self._get_curve_by_name(curve, ctype='EC')
+        return ec.EllipticCurvePublicNumbers(x, y, curve_fn)
+
+    def _ec_pri_n(self, curve):
+        d = self._decode_int(self.get('d'))
+        return ec.EllipticCurvePrivateNumbers(d, self._ec_pub_n(curve))
+
+    def _ec_pub(self, curve):
+        k = self._cache_pub_k
+        if k is None:
+            k = self._ec_pub_n(curve).public_key(default_backend())
+            self._cache_pub_k = k
+        return k
 
     def _ec_pri(self, curve):
-        d = self._decode_int(self.get('d'))
-        return ec.EllipticCurvePrivateNumbers(d, self._ec_pub(curve))
+        k = self._cache_pri_k
+        if k is None:
+            k = self._ec_pri_n(curve).private_key(default_backend())
+            self._cache_pri_k = k
+        return k
 
     def _okp_pub(self):
-        crv = self.get('crv')
-        try:
-            pubkey = _OKP_CURVES_TABLE[crv].pubkey
-        except KeyError:
-            raise InvalidJWKValue('Unknown curve "%s"' % crv)
+        k = self._cache_pub_k
+        if k is None:
+            crv = self.get('crv')
+            try:
+                pubkey = _OKP_CURVES_TABLE[crv].pubkey
+            except KeyError as e:
+                raise InvalidJWKValue('Unknown curve "%s"' % crv) from e
 
-        x = base64url_decode(self.get('x'))
-        return pubkey.from_public_bytes(x)
+            x = base64url_decode(self.get('x'))
+            k = pubkey.from_public_bytes(x)
+            self._cache_pub_k = k
+        return k
 
     def _okp_pri(self):
-        crv = self.get('crv')
-        try:
-            privkey = _OKP_CURVES_TABLE[crv].privkey
-        except KeyError:
-            raise InvalidJWKValue('Unknown curve "%s"' % crv)
+        k = self._cache_pri_k
+        if k is None:
+            crv = self.get('crv')
+            try:
+                privkey = _OKP_CURVES_TABLE[crv].privkey
+            except KeyError as e:
+                raise InvalidJWKValue('Unknown curve "%s"' % crv) from e
 
-        d = base64url_decode(self.get('d'))
-        return privkey.from_private_bytes(d)
+            d = base64url_decode(self.get('d'))
+            k = privkey.from_private_bytes(d)
+            self._cache_pri_k = k
+        return k
 
     def _get_public_key(self, arg=None):
         ktype = self.get('kty')
         if ktype == 'oct':
             return self.get('k')
         elif ktype == 'RSA':
-            return self._rsa_pub().public_key(default_backend())
+            return self._rsa_pub()
         elif ktype == 'EC':
-            return self._ec_pub(arg).public_key(default_backend())
+            return self._ec_pub(arg)
         elif ktype == 'OKP':
             return self._okp_pub()
         else:
@@ -795,9 +852,9 @@ class JWK(dict):
         if ktype == 'oct':
             return self.get('k')
         elif ktype == 'RSA':
-            return self._rsa_pri().private_key(default_backend())
+            return self._rsa_pri()
         elif ktype == 'EC':
-            return self._ec_pri(arg).private_key(default_backend())
+            return self._ec_pri(arg)
         elif ktype == 'OKP':
             return self._okp_pri()
         else:
@@ -811,7 +868,7 @@ class JWK(dict):
         :param operation: The requested operation.
          The valid set of operations is available in the
          :data:`JWKOperationsRegistry` registry.
-        :param arg: an optional, context specific, argument
+        :param arg: An optional, context specific, argument.
          For example a curve name.
 
         :raises InvalidJWKOperation: if the operation is unknown or
@@ -883,6 +940,7 @@ class JWK(dict):
                         data, backend=default_backend())
                     key = cert.public_key()
                 except ValueError:
+                    # pylint: disable=raise-missing-from
                     raise e
 
         self.import_from_pyca(key)
@@ -950,7 +1008,7 @@ class JWK(dict):
         """
 
         t = {'kty': self.get('kty')}
-        for name, val in iteritems(JWKValuesRegistry[t['kty']]):
+        for name, val in JWKValuesRegistry[t['kty']].items():
             if val.required:
                 t[name] = self.get(name)
         digest = hashes.Hash(hashalg, backend=default_backend())
@@ -972,6 +1030,9 @@ class JWK(dict):
 
         # Check if item is a key value and verify its format
         if item in list(JWKValuesRegistry[kty].keys()):
+            # Invalidate cached keys if any
+            self._cache_pub_k = None
+            self._cache_pri_k = None
             if JWKValuesRegistry[kty][item].type == ParmType.b64:
                 try:
                     v = base64url_decode(value)
@@ -980,17 +1041,17 @@ class JWK(dict):
                     # is used to indicate a 'None' key
                     if v == b'' and kty != 'oct' and item != 'k':
                         raise ValueError
-                except Exception:  # pylint: disable=broad-except
+                except Exception as e:  # pylint: disable=broad-except
                     raise InvalidJWKValue(
                         '"%s" is not base64url encoded' % item
-                    )
+                    ) from e
             elif JWKValuesRegistry[kty][item].type == ParmType.b64u:
                 try:
                     self._decode_int(value)
-                except Exception:  # pylint: disable=broad-except
+                except Exception as e:  # pylint: disable=broad-except
                     raise InvalidJWKValue(
                         '"%s" is not Base64urlUInt encoded' % item
-                    )
+                    ) from e
             super(JWK, self).__setitem__(item, value)
             return
 
@@ -1013,7 +1074,7 @@ class JWK(dict):
         super(JWK, self).__setitem__(item, value)
 
     def update(self, *args, **kwargs):
-        for k, v in iteritems(dict(*args, **kwargs)):
+        for k, v in dict(*args, **kwargs).items():
             self.__setitem__(k, v)
 
     def setdefault(self, key, default=None):
@@ -1031,6 +1092,12 @@ class JWK(dict):
                 if self.get(name) is not None:
                     raise KeyError("Cannot remove 'kty', values present")
 
+        kty = self.get('kty')
+        if kty is not None and item in list(JWKValuesRegistry[kty].keys()):
+            # Invalidate cached keys if any
+            self._cache_pub_k = None
+            self._cache_pri_k = None
+
         super(JWK, self).__delitem__(item)
 
     def __eq__(self, other):
@@ -1055,7 +1122,7 @@ class JWK(dict):
                         return self.get(item)
             raise KeyError
         except KeyError:
-            raise AttributeError
+            raise AttributeError(item) from None
 
     def __setattr__(self, item, value):
         try:
@@ -1066,26 +1133,26 @@ class JWK(dict):
                     self.__setitem__(item, value)
             super(JWK, self).__setattr__(item, value)
         except KeyError:
-            raise AttributeError
+            raise AttributeError(item) from None
 
     @classmethod
     def from_password(cls, password):
         """Creates a symmetric JWK key from a user password.
 
-        :param key: A password in utf8 format.
+        :param password: A password in utf8 format.
         """
         obj = cls()
         params = {'kty': 'oct'}
         try:
             params['k'] = base64url_encode(password.encode('utf8'))
         except Exception as e:  # pylint: disable=broad-except
-            raise InvalidJWKValue(e)
+            raise InvalidJWKValue from e
         obj.import_key(**params)
         return obj
 
     # Prevent accidental disclosure of key material via repr()
     def __repr__(self):
-        repr_dict = dict()
+        repr_dict = {}
         repr_dict['kid'] = self.get('kid', 'Missing Key ID')
         repr_dict['thumbprint'] = self.thumbprint()
         return json_encode(repr_dict)
@@ -1130,7 +1197,7 @@ class JWKSet(dict):
             super(JWKSet, self).__setitem__(key, val)
 
     def update(self, *args, **kwargs):
-        for k, v in iteritems(dict(*args, **kwargs)):
+        for k, v in dict(*args, **kwargs).items():
             self.__setitem__(k, v)
 
     def setdefault(self, key, default=None):
@@ -1150,10 +1217,10 @@ class JWKSet(dict):
         :param as_dict(bool): Whether to return a dict instead of
                               a JSON object
         """
-        exp_dict = dict()
-        for k, v in iteritems(self):
+        exp_dict = {}
+        for k, v in self.items():
             if k == 'keys':
-                keys = list()
+                keys = []
                 for jwk in v:
                     keys.append(jwk.export(private_keys, as_dict=True))
                 v = keys
@@ -1169,13 +1236,13 @@ class JWKSet(dict):
         """
         try:
             jwkset = json_decode(keyset)
-        except Exception:  # pylint: disable=broad-except
-            raise InvalidJWKValue()
+        except Exception as e:  # pylint: disable=broad-except
+            raise InvalidJWKValue from e
 
         if 'keys' not in jwkset:
-            raise InvalidJWKValue()
+            raise InvalidJWKValue
 
-        for k, v in iteritems(jwkset):
+        for k, v in jwkset.items():
             if k == 'keys':
                 for jwk in v:
                     self['keys'].add(JWK(**jwk))
@@ -1196,16 +1263,26 @@ class JWKSet(dict):
         """Gets a key from the set.
         :param kid: the 'kid' key identifier.
         """
-        for jwk in self['keys']:
-            if jwk.get('kid') == kid:
-                return jwk
-        return None
+        keys = self.get_keys(kid)
+        if len(keys) > 1:
+            raise InvalidJWKValue(
+                'Duplicate keys found with requested kid: 1 expected')
+        try:
+            return tuple(keys)[0]
+        except IndexError:
+            return None
+
+    def get_keys(self, kid):
+        """Gets keys from the set with matching kid.
+        :param kid: the 'kid' key identifier.
+        """
+        return {key for key in self['keys'] if key.get('kid') == kid}
 
     def __repr__(self):
-        repr_dict = dict()
-        for k, v in iteritems(self):
+        repr_dict = {}
+        for k, v in self.items():
             if k == 'keys':
-                keys = list()
+                keys = []
                 for jwk in v:
                     keys.append(repr(jwk))
                 v = keys
