@@ -85,7 +85,7 @@ class InvalidJWSOperation(JWException):
         super(InvalidJWSOperation, self).__init__(msg)
 
 
-class JWSCore(object):
+class JWSCore:
     """The inner JWS Core object.
 
     This object SHOULD NOT be used directly, the JWS object should be
@@ -108,6 +108,7 @@ class JWSCore(object):
         :raises ValueError: if the key is not a :class:`JWK` object
         :raises InvalidJWAAlgorithm: if the algorithm is not valid, is
             unknown or otherwise not yet implemented.
+        :raises InvalidJWSOperation: if the algorithm is not allowed.
         """
         self.alg = alg
         self.engine = self._jwa(alg, algs)
@@ -124,7 +125,7 @@ class JWSCore(object):
 
             self.protected = base64url_encode(header.encode('utf-8'))
         else:
-            self.header = dict()
+            self.header = {}
             self.protected = ''
         self.payload = payload
 
@@ -167,7 +168,7 @@ class JWSCore(object):
         return True
 
 
-class JWS(object):
+class JWS:
     """JSON Web Signature object
 
     This object represent a JWS token.
@@ -179,7 +180,7 @@ class JWS(object):
         :param payload(bytes): An arbitrary value (optional).
         :param header_registry: Optional additions to the header registry
         """
-        self.objects = dict()
+        self.objects = {}
         self.objects['payload'] = payload
         self.verifylog = None
         self._allowed_algs = None
@@ -236,7 +237,7 @@ class JWS(object):
             if hn is None:
                 continue
             if header is None:
-                header = dict()
+                header = {}
             for h in list(hn.keys()):
                 if h in self.header_registry:
                     if self.header_registry[h].mustprotect:
@@ -253,7 +254,7 @@ class JWS(object):
 
     # TODO: support selecting key with 'kid' and passing in multiple keys
     def _verify(self, alg, key, payload, signature, protected, header=None):
-        p = dict()
+        p = {}
         # verify it is a valid JSON object and decode
         if protected is not None:
             p = json_decode(protected)
@@ -288,23 +289,39 @@ class JWS(object):
         c = JWSCore(a, key, protected, payload, self._allowed_algs)
         c.verify(signature)
 
-    def verify(self, key, alg=None):
+    # Helper to deal with detached payloads in verification
+    def _get_obj_payload(self, obj, dp):
+        op = obj.get('payload')
+        if dp is not None:
+            if op is None or len(op) == 0:
+                return dp
+            else:
+                raise InvalidJWSOperation('Object Payload present but'
+                                          ' Deatched Payload provided')
+        return op
+
+    def verify(self, key, alg=None, detached_payload=None):
         """Verifies a JWS token.
 
         :param key: The (:class:`jwcrypto.jwk.JWK`) verification key.
-        :param alg: The signing algorithm (optional). usually the algorithm
+        :param alg: The signing algorithm (optional). Usually the algorithm
             is known as it is provided with the JOSE Headers of the token.
+        :param detached_payload: A detached payload to verify the signature
+            against. Only valid for tokens that are not carrying a payload.
 
         :raises InvalidJWSSignature: if the verification fails.
+        :raises InvalidJWSOperation: if a detached_payload is provided but
+                                     an object payload exists
         """
 
-        self.verifylog = list()
+        self.verifylog = []
         self.objects['valid'] = False
         obj = self.objects
         if 'signature' in obj:
             try:
+                payload = self._get_obj_payload(obj, detached_payload)
                 self._verify(alg, key,
-                             obj['payload'],
+                             payload,
                              obj['signature'],
                              obj.get('protected', None),
                              obj.get('header', None))
@@ -315,8 +332,9 @@ class JWS(object):
         elif 'signatures' in obj:
             for o in obj['signatures']:
                 try:
+                    payload = self._get_obj_payload(obj, detached_payload)
                     self._verify(alg, key,
-                                 obj['payload'],
+                                 payload,
                                  o['signature'],
                                  o.get('protected', None),
                                  o.get('header', None))
@@ -332,8 +350,7 @@ class JWS(object):
                                       'signatures' + repr(self.verifylog))
 
     def _deserialize_signature(self, s):
-        o = dict()
-        o['signature'] = base64url_decode(str(s['signature']))
+        o = {'signature': base64url_decode(str(s['signature']))}
         if 'protected' in s:
             p = base64url_decode(str(s['protected']))
             o['protected'] = p.decode('utf-8')
@@ -369,19 +386,19 @@ class JWS(object):
         :param key: A (:class:`jwcrypto.jwk.JWK`) verification key (optional).
          If a key is provided a verification step will be attempted after
          the object is successfully deserialized.
-        :param alg: The signing algorithm (optional). usually the algorithm
+        :param alg: The signing algorithm (optional). Usually the algorithm
          is known as it is provided with the JOSE Headers of the token.
 
         :raises InvalidJWSObject: if the raw object is an invalid JWS token.
         :raises InvalidJWSSignature: if the verification fails.
         """
-        self.objects = dict()
-        o = dict()
+        self.objects = {}
+        o = {}
         try:
             try:
                 djws = json_decode(raw_jws)
                 if 'signatures' in djws:
-                    o['signatures'] = list()
+                    o['signatures'] = []
                     for s in djws['signatures']:
                         os = self._deserialize_signature(s)
                         o['signatures'].append(os)
@@ -437,7 +454,7 @@ class JWS(object):
 
         b64 = True
 
-        p = dict()
+        p = {}
         if protected:
             if isinstance(protected, dict):
                 p = protected
@@ -482,20 +499,20 @@ class JWS(object):
         )
         sig = c.sign()
 
-        o = dict()
-        o['signature'] = base64url_decode(sig['signature'])
+        o = {
+            'signature': base64url_decode(sig['signature']),
+            'valid': True,
+        }
         if protected:
             o['protected'] = protected
         if header:
             o['header'] = h
-        o['valid'] = True
 
         if 'signatures' in self.objects:
             self.objects['signatures'].append(o)
         elif 'signature' in self.objects:
-            self.objects['signatures'] = list()
-            n = dict()
-            n['signature'] = self.objects.pop('signature')
+            self.objects['signatures'] = []
+            n = {'signature': self.objects.pop('signature')}
             if 'protected' in self.objects:
                 n['protected'] = self.objects.pop('protected')
             if 'header' in self.objects:
@@ -554,7 +571,7 @@ class JWS(object):
                              base64url_encode(self.objects['signature'])])
         else:
             obj = self.objects
-            sig = dict()
+            sig = {}
             payload = self.objects.get('payload', '')
             if self.objects.get('b64', True):
                 sig['payload'] = base64url_encode(payload)
@@ -569,7 +586,7 @@ class JWS(object):
                 if 'header' in obj:
                     sig['header'] = obj['header']
             elif 'signatures' in obj:
-                sig['signatures'] = list()
+                sig['signatures'] = []
                 for o in obj['signatures']:
                     if not o.get('valid', False):
                         continue
@@ -602,16 +619,16 @@ class JWS(object):
                 p = json_decode(obj['protected'])
             else:
                 p = None
-            return self._merge_check_headers(p, obj.get('header', dict()))
+            return self._merge_check_headers(p, obj.get('header', {}))
         elif 'signatures' in self.objects:
-            jhl = list()
+            jhl = []
             for o in obj['signatures']:
-                jh = dict()
+                jh = {}
                 if 'protected' in o:
                     p = json_decode(o['protected'])
                 else:
                     p = None
-                jh = self._merge_check_headers(p, o.get('header', dict()))
+                jh = self._merge_check_headers(p, o.get('header', {}))
                 jhl.append(jh)
             return jhl
         else:
