@@ -237,7 +237,7 @@ zl9HYIMxATFyqSiD9jsx
 -----END CERTIFICATE-----
 """
 
-PublicCertThumbprint = u'7KITkGJF74IZ9NKVvHfuJILbuIZny6j-roaNjB1vgiA'
+PublicCertThumbprint = '7KITkGJF74IZ9NKVvHfuJILbuIZny6j-roaNjB1vgiA'
 
 # RFC 8037 - A.2
 PublicKeys_EdDsa = {
@@ -269,7 +269,13 @@ PublicKeys_secp256k1 = {
             "crv": "secp256k1",
             "x": "Ss6na3mcci8Ud4lQrjaB_T40sfKApEcl2RLIWOJdjow",
             "y": "7l9qIKtKPW6oEiOYBt7r22Sm0mtFJU-yBkkvMvpscd8"
-        }
+        },
+        {
+            "kty": "EC",
+            "crv": "P-256K",
+            "x": "Ss6na3mcci8Ud4lQrjaB_T40sfKApEcl2RLIWOJdjow",
+            "y": "7l9qIKtKPW6oEiOYBt7r22Sm0mtFJU-yBkkvMvpscd8"
+        },
     ]
 }
 
@@ -282,6 +288,13 @@ PrivateKeys_secp256k1 = {
             "y": "7l9qIKtKPW6oEiOYBt7r22Sm0mtFJU-yBkkvMvpscd8",
             "d": "GYhU2vrYGZrjLZn71Xniqm54Mi53xiYtaTLawzaf9dA"
         },
+        {
+            "kty": "EC",
+            "crv": "P-256K",
+            "x": "Ss6na3mcci8Ud4lQrjaB_T40sfKApEcl2RLIWOJdjow",
+            "y": "7l9qIKtKPW6oEiOYBt7r22Sm0mtFJU-yBkkvMvpscd8",
+            "d": "GYhU2vrYGZrjLZn71Xniqm54Mi53xiYtaTLawzaf9dA"
+        },
     ]
 }
 
@@ -362,21 +375,21 @@ class TestJWK(unittest.TestCase):
     def test_generate_EC_key(self):
         # Backwards compat curve
         key = jwk.JWK.generate(kty='EC', curve='P-256')
-        key.get_curve('P-256')
+        key.get_op_key('verify', 'P-256')
         # New param
         key = jwk.JWK.generate(kty='EC', crv='P-521')
-        key.get_curve('P-521')
+        key.get_op_key('verify', 'P-521')
         # New param prevails
         key = jwk.JWK.generate(kty='EC', curve='P-256', crv='P-521')
-        key.get_curve('P-521')
+        key.get_op_key('verify', 'P-521')
         # New secp256k curve
         key = jwk.JWK.generate(kty='EC', curve='secp256k1')
-        key.get_curve('secp256k1')
+        key.get_op_key('verify', 'secp256k1')
 
     def test_generate_OKP_keys(self):
         for crv in jwk.ImplementedOkpCurves:
             key = jwk.JWK.generate(kty='OKP', crv=crv)
-            self.assertEqual(key.get_curve(crv), crv)
+            self.assertEqual(key['crv'], crv)
 
     def test_import_pyca_keys(self):
         rsa1 = rsa.generate_private_key(65537, 1024, default_backend())
@@ -392,7 +405,7 @@ class TestJWK(unittest.TestCase):
         self.assertEqual(kec1.get_op_key('verify').public_numbers().x,
                          kec2.get_op_key('verify').public_numbers().x)
         self.assertRaises(jwk.InvalidJWKValue,
-                          jwk.JWK.from_pyca, dict())
+                          jwk.JWK.from_pyca, {})
 
     def test_jwk_from_json(self):
         k = jwk.JWK.generate(kty='oct', size=256)
@@ -424,6 +437,24 @@ class TestJWK(unittest.TestCase):
             num += 1
         self.assertEqual(num, len(PrivateKeys['keys']))
 
+    def test_jwkset_get_keys(self):
+        # Test key set with mutiple keys
+        ksm = jwk.JWKSet.from_json(json_encode(PrivateKeys))
+        k1 = jwk.JWK.from_json(json_encode(PrivateKeys['keys'][0]))
+        kwargs = RSAPrivateKey.copy()
+        kwargs['kid'] = '1'
+        k2 = jwk.JWK(**kwargs)
+        self.assertEqual(k1, ksm.get_key('1'))
+        self.assertIsNone(ksm.get_key('not-there'))
+
+        ksm.add(k2)
+        self.assertEqual({k1, k2}, ksm.get_keys('1'))
+        self.assertEqual(3, len(ksm['keys']))
+        # Expect that duplicate kids will
+        # raise an exception when we use get_key
+        with self.assertRaises(jwk.InvalidJWKValue):
+            ksm.get_key('1')
+
     def test_jwkset_issue_208(self):
         ks = jwk.JWKSet()
         key1 = RSAPrivateKey.copy()
@@ -501,7 +532,7 @@ class TestJWK(unittest.TestCase):
         key = jwk.JWK.from_pem(RSAPrivatePEM, password=RSAPrivatePassword)
         k = key.export_private(as_dict=True)
         self.assertEqual(k['kid'],
-                         u'x31vrbZceU2qOPLtrUwPkLa3PNakMn9tOsq_ntFVrJc')
+                         'x31vrbZceU2qOPLtrUwPkLa3PNakMn9tOsq_ntFVrJc')
         keyset = jwk.JWKSet.from_json(json_encode(PrivateKeys))
         ks = keyset.export(as_dict=True)
         self.assertTrue('keys' in ks)
@@ -595,6 +626,21 @@ class TestJWK(unittest.TestCase):
         self.assertEqual(key['kty'], 'oct')
         self.assertEqual(key['k'], 'dGVzdCBwYXNzd29yZA')
 
+    def test_p256k_alias(self):
+        key = jwk.JWK.generate(kty='EC', curve='P-256K')
+        key.get_op_key('verify', 'secp256k1')
+
+        pub_k = jwk.JWK(**PrivateKeys_secp256k1['keys'][0])
+        pri_k = jwk.JWK(**PrivateKeys_secp256k1['keys'][1])
+        payload = bytes(bytearray(A1_payload))
+        test = jws.JWS(payload)
+        test.add_signature(pri_k, None, json_encode({"alg": "ES256K"}), None)
+        test_serialization_compact = test.serialize(compact=True)
+        verify = jws.JWS()
+        verify.deserialize(test_serialization_compact)
+        verify.verify(pub_k.public())
+        self.assertEqual(verify.payload, payload)
+
 
 # RFC 7515 - A.1
 A1_protected = \
@@ -934,6 +980,21 @@ class TestJWS(unittest.TestCase):
         jws_verify.verify(key.public())
         self.assertEqual(jws_verify.payload, payload)
 
+    def test_jws_issue_224(self):
+        key = jwk.JWK().generate(kty='oct')
+
+        # Test Empty payload is supported for creating and verifying signatures
+        s = jws.JWS(payload='')
+        s.add_signature(key, None, json_encode({"alg": "HS256"}))
+        o1 = s.serialize(compact=True)
+        self.assertTrue('..' in o1)
+        o2 = json_decode(s.serialize())
+        self.assertEqual(o2['payload'], '')
+
+        t = jws.JWS()
+        t.deserialize(o1)
+        t.verify(key)
+
 
 E_A1_plaintext = \
     [84, 104, 101, 32, 116, 114, 117, 101, 32, 115, 105, 103, 110, 32,
@@ -1427,7 +1488,9 @@ class TestJWT(unittest.TestCase):
         t.make_encrypted_token(key)
         token = t.serialize()
         # try to decrypt without a matching key
-        self.assertRaises(jwt.JWTMissingKey, jwt.JWT, jwt=token, key=keyset)
+        self.assertRaises(jwt.JWTMissingKey, jwt.JWT, jwt=token, key=keyset,
+                          algs=jwe_algs_and_rsa1_5,
+                          check_claims={'exp': 1300819380})
         # now decrypt with key
         keyset.add(key)
         jwt.JWT(jwt=token, key=keyset, algs=jwe_algs_and_rsa1_5,
@@ -1439,7 +1502,7 @@ class TestJWT(unittest.TestCase):
         t = jwt.JWT(header, A1_claims, algs=jwe_algs_and_rsa1_5)
         t.make_encrypted_token(key)
         token = t.serialize()
-        self.assertRaises(jwe.InvalidJWEData, jwt.JWT, jwt=token, key=keyset)
+        self.assertRaises(jwt.JWTMissingKey, jwt.JWT, jwt=token, key=keyset)
 
         keyset = jwk.JWKSet.from_json(json_encode(PrivateKeys))
         # encrypt a new JWT with no kid
@@ -1448,12 +1511,40 @@ class TestJWT(unittest.TestCase):
         t.make_encrypted_token(key)
         token = t.serialize()
         # try to decrypt without a matching key
-        self.assertRaises(jwt.JWTMissingKey, jwt.JWT, jwt=token, key=keyset)
+        self.assertRaises(jwt.JWTMissingKey, jwt.JWT, jwt=token, key=keyset,
+                          algs=jwe_algs_and_rsa1_5,
+                          check_claims={'exp': 1300819380})
         # now decrypt with key
         keyset.add(key)
         jwt.JWT(jwt=token, key=keyset, algs=jwe_algs_and_rsa1_5,
                 check_claims={'exp': 1300819380})
 
+    def test_decrypt_keyset_dup_kid(self):
+        keyset = jwk.JWKSet.from_json(json_encode(PrivateKeys))
+        # add wrong key with duplicate kid
+        key = jwk.JWK(kid='testkey', **E_A3_key)
+        keyset.add(key)
+
+        # encrypt a new JWT with kid
+        key = jwk.JWK(kid='testkey', **E_A2_key)
+        header = copy.copy(A1_header)
+        header['kid'] = 'testkey'
+        t = jwt.JWT(header, A1_claims, algs=jwe_algs_and_rsa1_5)
+        t.make_encrypted_token(key)
+        token = t.serialize()
+
+        # try to decrypt without a matching key
+        with self.assertRaises(jwt.JWTMissingKey):
+            jwt.JWT(jwt=token, key=keyset, algs=jwe_algs_and_rsa1_5,
+                    check_claims={'exp': 1300819380})
+
+        # add right key
+        keyset.add(key)
+
+        # now decrypt with key
+        jwt.JWT(jwt=token, key=keyset, algs=jwe_algs_and_rsa1_5,
+                check_claims={'exp': 1300819380})
+
     def test_invalid_claim_type(self):
         key = jwk.JWK(**E_A2_key)
         claims = {"testclaim": "test"}
@@ -1492,6 +1583,38 @@ class TestJWT(unittest.TestCase):
                 check_claims={"iss": "test", "exp": None,
                               "string_claim": "test"})
 
+    def test_claims_typ(self):
+        key = jwk.JWK().generate(kty='oct')
+        claims = '{"typ":"application/test"}'
+        string_header = '{"alg":"HS256"}'
+        t = jwt.JWT(string_header, claims)
+        t.make_signed_token(key)
+        token = t.serialize()
+
+        # Same typ w/o application prefix
+        jwt.JWT(jwt=token, key=key, check_claims={"typ": "test"})
+        self.assertRaises(jwt.JWTInvalidClaimValue, jwt.JWT, jwt=token,
+                          key=key, check_claims={"typ": "wrong"})
+
+        # Same typ w/ application prefix
+        jwt.JWT(jwt=token, key=key, check_claims={"typ": "application/test"})
+        self.assertRaises(jwt.JWTInvalidClaimValue, jwt.JWT, jwt=token,
+                          key=key, check_claims={"typ": "application/wrong"})
+
+        # check that a '/' in the name makes it not be matched with
+        # 'application/' prefix
+        claims = '{"typ":"diffmime/test"}'
+        t = jwt.JWT(string_header, claims)
+        t.make_signed_token(key)
+        token = t.serialize()
+        self.assertRaises(jwt.JWTInvalidClaimValue, jwt.JWT, jwt=token,
+                          key=key, check_claims={"typ": "application/test"})
+        self.assertRaises(jwt.JWTInvalidClaimValue, jwt.JWT, jwt=token,
+                          key=key, check_claims={"typ": "test"})
+
+        # finally make sure it doesn't raise if not checked.
+        jwt.JWT(jwt=token, key=key)
+
     def test_empty_claims(self):
         key = jwk.JWK().generate(kty='oct')
 
@@ -1505,12 +1628,12 @@ class TestJWT(unittest.TestCase):
         c.deserialize(token, key)
         self.assertEqual('{}', c.claims)
 
-        # empty string is not valid
+        # empty string is also valid
         t = jwt.JWT('{"alg":"HS256"}', '')
-        self.assertEqual('', t.claims)
-        self.assertRaises(jws.InvalidJWSObject, t.make_signed_token, key)
+        t.make_signed_token(key)
+        token = t.serialize()
 
-        # but a space is fine
+        # also a space is fine
         t = jwt.JWT('{"alg":"HS256"}', ' ')
         self.assertEqual(' ', t.claims)
         t.make_signed_token(key)
@@ -1534,6 +1657,15 @@ class TestJWT(unittest.TestCase):
         # the oct key before hitting the ES one
         jwt.JWT(jwt=token, key=ks)
 
+    def test_Issue_239(self):
+        claims = {"aud": "www.example.com"}
+        check_claims = {"aud": ["www.example.com", "account"]}
+        key = jwk.JWK(generate='oct', size=256)
+        token = jwt.JWT(header={"alg": "HS256"}, claims=claims)
+        token.make_signed_token(key)
+        self.assertRaises(jwt.JWTInvalidClaimFormat, jwt.JWT, key=key,
+                          jwt=token.serialize(), check_claims=check_claims)
+
 
 class ConformanceTests(unittest.TestCase):
 
@@ -1640,8 +1772,8 @@ class ConformanceTests(unittest.TestCase):
 
     def test_no_default_rsa_1_5(self):
         s = jws.JWS('test')
-        with self.assertRaisesRegexp(jws.InvalidJWSOperation,
-                                     'Algorithm not allowed'):
+        with self.assertRaisesRegex(jws.InvalidJWSOperation,
+                                    'Algorithm not allowed'):
             s.add_signature(A2_key, alg="RSA1_5")
 
     def test_pbes2_hs256_aeskw(self):
@@ -1737,6 +1869,17 @@ class TestUnencodedPayload(unittest.TestCase):
         sig = s.serialize(compact=True)
         self.assertEqual(sig, result)
 
+    def test_detached_payload_verification(self):
+        token = \
+            'eyJhbGciOiJIUzI1NiIsImI2NCI6ZmFsc2UsImNyaXQiOlsiYjY0Il19..' + \
+            'A5dxf2s96_n5FLueVuW1Z_vh161FwXZC4YLPff6dmDY'
+
+        s = jws.JWS()
+        s.deserialize(token)
+        s.verify(jwk.JWK(**SymmetricKeys['keys'][1]),
+                 detached_payload=rfc7797_payload)
+        self.assertTrue(s.is_valid)
+
     def test_misses_crit(self):
         s = jws.JWS(rfc7797_payload)
         with self.assertRaises(jws.InvalidJWSObject):
